require "defines"

-- Has to be a power of 2 to work correctly
local itemCollisionDistance = 0.5
local droppedItemName = "item-on-ground"
local dummyDroppedItemName = "dummy-item-on-ground"
local spillableChests = {
	-- Vanilla chests
	["wooden-chest"] = true,
	["iron-chest"] = true,
	["steel-chest"] = true,
	-- Vanilla logicstics chests
	["smart-chest"] = true,
	["logistic-chest-passive-provider"] = true,
	["logistic-chest-active-provider"] = true,
	["logistic-chest-storage"] = true,
	["logistic-chest-requester"] = true,
	-- Dytech standard chests
	["wooden-chest-medium"] = true,
	["iron-chest-medium"] = true,
	["steel-chest-medium"] = true,
	["wooden-chest-big"] = true,
	["iron-chest-big"] = true,
	["steel-chest-big"] = true,
	-- Dytech logistics chests
	["logistic-chest-passive-provider-1"] = true,
	["logistic-chest-active-provider-1"] = true,
	["logistic-chest-storage-1"] = true,
	["logistic-chest-requester-1"] = true,
	["logistic-chest-passive-provider-2"] = true,
	["logistic-chest-active-provider-2"] = true,
	["logistic-chest-storage-2"] = true,
	["logistic-chest-requester-2"] = true
}

function canInsert(_entity, _index, _stack)
	-- NOTE: if checking the hotbar and the player has their hotbar filtered and there's nothing in the filtered slot caninsert will return true for that slot
	local stackSize = game.itemprototypes[_stack.name].stacksize
	local results = {["count"] = 0, ["slots"] = 0}
	local toInsertCount = _stack.count
	
	if toInsertCount ~= 0 then
		for _,index in pairs(_index) do
			inventoryContents = _entity.getinventory(index)
			
			for slot = 1,#inventoryContents do
				contents = inventoryContents[slot]
				
				if contents ~= nil then
					-- Slot with contents
					if contents.name == _stack.name then
						if contents.count < stackSize then
							if toInsertCount >= stackSize - contents.count then
								results["count"] = results["count"] + (stackSize - contents.count)
								toInsertCount = toInsertCount - (stackSize - contents.count)
							else
								results["count"] = results["count"] + toInsertCount
								toInsertCount = 0
								break
							end
						end
					end
				else
					-- Empty slot
					results["slots"] = results["slots"] + 1
					
					if toInsertCount >= stackSize then
						results["count"] = results["count"] + stackSize
						toInsertCount = toInsertCount - stackSize
					else
						results["count"] = results["count"] + toInsertCount
						toInsertCount = 0
						break
					end
				end
			end
			
			if toInsertCount == 0 then
				break
			end
		end
	end
	
	return results
end


game.onevent(defines.events.onpreplayermineditem, function(event)
	if spillableChests[event.entity.name] == true then
		local chestContents = event.entity.getinventory(1).getcontents()
		local chestHasContents
		
		for k,v in pairs(chestContents) do
			chestHasContents = true
			break
		end
		
		if chestHasContents then
			local position = {x = event.entity.position.x + itemCollisionDistance, y = event.entity.position.y - itemCollisionDistance}
			local countBefore
			local countAfter
			local spillCount
			local retry
			local side = 1
			local cornerX = event.entity.position.x - itemCollisionDistance
			local cornerY = event.entity.position.y - itemCollisionDistance
			local squareSSLength = itemCollisionDistance + itemCollisionDistance
			local step
			local availablePlayerSlots = 0
			local playerInventory = game.player.getinventory(2)
			local results
			local stackSize
			
			
			for n = 1, #playerInventory do
				if playerInventory[n] == nil then
					availablePlayerSlots = availablePlayerSlots + 1
				end
			end
			
			for item,count in pairs(chestContents) do
				if item ~= "blueprint" then
					results = canInsert(game.player, {[1] = 2}, {name = item, count = count})
					
					if results["count"] > 0 then
						if results["slots"] > 0 then
							stackSize = game.itemprototypes[item].stacksize
							
							if availablePlayerSlots > 0 then
								if availablePlayerSlots - results["slots"] >= 0 then
									availablePlayerSlots = availablePlayerSlots - results["slots"]
									spillCount = count - results["count"]
								else
									spillCount = count - (results["count"] - ((results["slots"] - availablePlayerSlots) * stackSize))
									availablePlayerSlots = 0
								end
							else
								spillCount = count - (results["count"] - (results["slots"] * stackSize))
							end
						else
							if results["count"] ~= count then
								spillCount = count - results["count"]
							else
								spillCount = 0
							end
						end
					else
						spillCount = count
					end
					
					if spillCount ~= 0 then
						for n = 1, spillCount do
							retry = true
							while retry do
								if game.canplaceentity({name = dummyDroppedItemName, position = position}) then
									game.createentity({name = droppedItemName, position = position, stack = {name = item, count = 1}})
									retry = false
								end
								
								step = true
								if side == 1 then
									if position.x <= cornerX then
										cornerY = cornerY + squareSSLength
										side = 2
									end
								elseif side == 2 then
									if position.y >= cornerY then
										cornerX = cornerX + squareSSLength
										side = 3
									end
								elseif side == 3 then
									if position.x >= cornerX then
										cornerY = (cornerY - squareSSLength) + itemCollisionDistance
										side = 4
									end
								elseif side == 4 then
									if position.y <= cornerY then
										step = false
										squareSSLength = squareSSLength + itemCollisionDistance + itemCollisionDistance
										side = 1
										
										position.x = position.x + itemCollisionDistance
										cornerX = position.x - squareSSLength
										position.y = (position.y - itemCollisionDistance) - itemCollisionDistance
										cornerY = position.y
									end
								end
								
								if step == true then
									if side == 1 then
										position.x = position.x - itemCollisionDistance
									elseif side == 2 then
										position.y = position.y + itemCollisionDistance
									elseif side == 3 then
										position.x = position.x + itemCollisionDistance
									elseif side == 4 then
										position.y = position.y - itemCollisionDistance
									end
								end
							end
						end
						
						event.entity.getinventory(1).remove({name = item, count = spillCount})
					end
				end
			end
		end
	end
end)